@@ -20,46 +20,7 @@
 #pragma interface			/* gcc class implementation */
 #endif
 
-#include "my_sys.h"                    /* alloc_root, TRASH, MY_WME,
-                                          MY_FAE, MY_ALLOW_ZERO_PTR */
-#include "m_string.h"                           /* bfill */
-
-THD *thd_get_current_thd();
-
-/* mysql standard class memory allocator */
-
-class Sql_alloc
-{
-public:
-  static void *operator new(size_t size) throw ()
-  {
-    return thd_alloc(thd_get_current_thd(), size);
-  }
-  static void *operator new[](size_t size) throw ()
-  {
-    return thd_alloc(thd_get_current_thd(), size);
-  }
-  static void *operator new[](size_t size, MEM_ROOT *mem_root) throw ()
-  { return alloc_root(mem_root, size); }
-  static void *operator new(size_t size, MEM_ROOT *mem_root) throw ()
-  { return alloc_root(mem_root, size); }
-  static void operator delete(void *ptr, size_t size) { TRASH_FREE(ptr, size); }
-  static void operator delete(void *ptr, MEM_ROOT *mem_root)
-  { /* never called */ }
-  static void operator delete[](void *ptr, MEM_ROOT *mem_root)
-  { /* never called */ }
-  static void operator delete[](void *ptr, size_t size) { TRASH_FREE(ptr, size); }
-#ifdef HAVE_valgrind
-  bool dummy_for_valgrind;
-  inline Sql_alloc() :dummy_for_valgrind(0) {}
-  inline ~Sql_alloc() {}
-#else
-  inline Sql_alloc() {}
-  inline ~Sql_alloc() {}
-#endif
-
-};
-
+#include "sql_alloc.h"
 
 /**
   Simple intrusive linked list.
@@ -68,6 +29,7 @@ public:
           a pointer to the first element in the list and a indirect
           reference to the last element.
 */
+
 template <typename T>
 class SQL_I_List :public Sql_alloc
 {
@@ -156,8 +118,8 @@ struct list_node :public Sql_alloc
 {
   list_node *next;
   void *info;
-  list_node(void *info_par,list_node *next_par)
-    :next(next_par),info(info_par)
+  list_node(const void *info_par, list_node *next_par)
+    :next(next_par), info(const_cast<void *>(info_par))
   {}
   list_node()					/* For end_of_list */
   {
@@ -352,16 +314,15 @@ public:
   friend class error_list;
   friend class error_list_iterator;
 
-#ifndef DBUG_OFF
   /*
-    Debugging help: return N-th element in the list, or NULL if the list has
+    Return N-th element in the list, or NULL if the list has
     less than N elements.
   */
-  void *elem(int n)
+  void *elem(uint n)
   {
     list_node *node= first;
     void *data= NULL;
-    for (int i=0; i <= n; i++)
+    for (uint i= 0; i <= n; i++)
     {
       if (node == &end_of_list)
       {
@@ -373,7 +334,6 @@ public:
     }
     return data;
   }
-#endif
 
 #ifdef LIST_EXTRA_DEBUG
   /*
@@ -424,7 +384,7 @@ public:
 #endif // LIST_EXTRA_DEBUG
 
 protected:
-  void after(void *info,list_node *node)
+  void after(const void *info, list_node *node)
   {
     list_node *new_node=new list_node(info,node->next);
     node->next=new_node;
@@ -485,11 +445,11 @@ public:
   {
     el= &list->first;
   }
-  inline void *replace(void *element)
+  inline void *replace(const void *element)
   {						// Return old element
     void *tmp=current->info;
     DBUG_ASSERT(current->info != 0);
-    current->info=element;
+    current->info= const_cast<void *>(element);
     return tmp;
   }
   void *replace(base_list &new_list)
@@ -512,7 +472,7 @@ public:
     el=prev;
     current=0;					// Safeguard
   }
-  void after(void *element)			// Insert element after current
+  void after(const void *element)		// Insert element after current
   {
     list->after(element,current);
     current=current->next;
@@ -541,10 +501,10 @@ public:
     base_list(tmp, mem_root) {}
   inline bool push_back(T *a) { return base_list::push_back(a); }
   inline bool push_back(T *a, MEM_ROOT *mem_root)
-  { return base_list::push_back(a, mem_root); }
+  { return base_list::push_back((void*) a, mem_root); }
   inline bool push_front(T *a) { return base_list::push_front(a); }
   inline bool push_front(T *a, MEM_ROOT *mem_root)
-  { return base_list::push_front(a, mem_root); }
+  { return base_list::push_front((void*) a, mem_root); }
   inline T* head() {return (T*) base_list::head(); }
   inline T** head_ref() {return (T**) base_list::head_ref(); }
   inline T* pop()  {return (T*) base_list::pop(); }
@@ -565,9 +525,7 @@ public:
     }
     empty();
   }
-#ifndef DBUG_OFF
-  T *elem(int n) { return (T*)base_list::elem(n); }
-#endif
+  T *elem(uint n) { return (T*) base_list::elem(n); }
 };
 
 
@@ -591,10 +549,10 @@ public:
 template <class T> class List_iterator_fast :public base_list_iterator
 {
 protected:
-  inline T *replace(T *a)   { return (T*) 0; }
-  inline T *replace(List<T> &a) { return (T*) 0; }
-  inline void remove(void)  { }
-  inline void after(T *a)   { }
+  inline T *replace(T *)   { return (T*) 0; }
+  inline T *replace(List<T> &) { return (T*) 0; }
+  inline void remove(void)  {}
+  inline void after(T *)    {}
   inline T** ref(void)	    { return (T**) 0; }
 
 public:
@@ -663,7 +621,7 @@ struct ilink
   {
     return (void*)my_malloc((uint)size, MYF(MY_WME | MY_FAE | ME_FATALERROR));
   }
-  static void operator delete(void* ptr_arg, size_t size)
+  static void operator delete(void* ptr_arg, size_t)
   {
      my_free(ptr_arg);
   }
@@ -683,6 +641,10 @@ struct ilink
   {
     DBUG_ASSERT(prev != 0 && next != 0);
   }
+  inline void assert_not_linked()
+  {
+    DBUG_ASSERT(prev == 0 && next == 0);
+  }
   virtual ~ilink() { unlink(); }		/*lint -e1740 */
 };
 
