@@ -2,7 +2,7 @@
 
 Copyright (c) 1996, 2016, Oracle and/or its affiliates. All Rights Reserved.
 Copyright (c) 2008, Google Inc.
-Copyright (c) 2017, 2021, MariaDB Corporation.
+Copyright (c) 2017, 2022, MariaDB Corporation.
 
 Portions of this file contain modifications contributed and copyrighted by
 Google, Inc. Those modifications are gratefully acknowledged and are described
@@ -45,10 +45,10 @@ Created 2/17/1996 Heikki Tuuri
 
 /** Is search system enabled.
 Search system is protected by array of latches. */
-char		btr_search_enabled	= true;
+char		btr_search_enabled;
 
 /** Number of adaptive hash index partition. */
-ulong		btr_ahi_parts		= 8;
+ulong		btr_ahi_parts;
 
 #ifdef UNIV_SEARCH_PERF_STAT
 /** Number of successful adaptive hash index lookups */
@@ -84,7 +84,74 @@ is then built on the page, assuming the global limit has been reached */
 
 /** The global limit for consecutive potentially successful hash searches,
 before hash index building is started */
-#define BTR_SEARCH_BUILD_LIMIT		100
+#define BTR_SEARCH_BUILD_LIMIT		100U
+
+/** Compute a hash value of a record in a page.
+@param[in]	rec		index record
+@param[in]	offsets		return value of rec_get_offsets()
+@param[in]	n_fields	number of complete fields to fold
+@param[in]	n_bytes		number of bytes to fold in the last field
+@param[in]	index_id	index tree ID
+@return the hash value */
+static inline
+ulint
+rec_fold(
+	const rec_t*	rec,
+	const rec_offs*	offsets,
+	ulint		n_fields,
+	ulint		n_bytes,
+	index_id_t	tree_id)
+{
+	ulint		i;
+	const byte*	data;
+	ulint		len;
+	ulint		fold;
+	ulint		n_fields_rec;
+
+	ut_ad(rec_offs_validate(rec, NULL, offsets));
+	ut_ad(rec_validate(rec, offsets));
+	ut_ad(page_rec_is_leaf(rec));
+	ut_ad(!page_rec_is_metadata(rec));
+	ut_ad(n_fields > 0 || n_bytes > 0);
+
+	n_fields_rec = rec_offs_n_fields(offsets);
+	ut_ad(n_fields <= n_fields_rec);
+	ut_ad(n_fields < n_fields_rec || n_bytes == 0);
+
+	if (n_fields > n_fields_rec) {
+		n_fields = n_fields_rec;
+	}
+
+	if (n_fields == n_fields_rec) {
+		n_bytes = 0;
+	}
+
+	fold = ut_fold_ull(tree_id);
+
+	for (i = 0; i < n_fields; i++) {
+		data = rec_get_nth_field(rec, offsets, i, &len);
+
+		if (len != UNIV_SQL_NULL) {
+			fold = ut_fold_ulint_pair(fold,
+						  ut_fold_binary(data, len));
+		}
+	}
+
+	if (n_bytes > 0) {
+		data = rec_get_nth_field(rec, offsets, i, &len);
+
+		if (len != UNIV_SQL_NULL) {
+			if (len > n_bytes) {
+				len = n_bytes;
+			}
+
+			fold = ut_fold_ulint_pair(fold,
+						  ut_fold_binary(data, len));
+		}
+	}
+
+	return(fold);
+}
 
 /** Determine the number of accessed key fields.
 @param[in]	n_fields	number of complete fields
@@ -110,23 +177,6 @@ btr_search_get_n_fields(
 	return(btr_search_get_n_fields(cursor->n_fields, cursor->n_bytes));
 }
 
-/********************************************************************//**
-Builds a hash index on a page with the given parameters. If the page already
-has a hash index with different parameters, the old hash index is removed.
-If index is non-NULL, this function checks if n_fields and n_bytes are
-sensible values, and does not build a hash index if not. */
-static
-void
-btr_search_build_page_hash_index(
-/*=============================*/
-	dict_index_t*	index,	/*!< in: index for which to build, or NULL if
-				not known */
-	buf_block_t*	block,	/*!< in: index page, s- or x-latched */
-	ulint		n_fields,/*!< in: hash this many full fields */
-	ulint		n_bytes,/*!< in: hash this many bytes from the next
-				field */
-	ibool		left_side);/*!< in: hash for searches from left side? */
-
 /** This function should be called before reserving any btr search mutex, if
 the intended operation might add nodes to the search system hash table.
 Because of the latching order, once we have reserved the btr search system
@@ -139,7 +189,7 @@ will not guarantee success.
 @param[in]	index	index handler */
 static
 void
-btr_search_check_free_space_in_heap(dict_index_t* index)
+btr_search_check_free_space_in_heap(const dict_index_t* index)
 {
 	/* Note that we peek the value of heap->free_block without reserving
 	the latch: this is ok, because we will not guarantee that there will
@@ -171,8 +221,7 @@ func_exit:
 
 /** Creates and initializes the adaptive search system at a database start.
 @param[in]	hash_size	hash table size. */
-void
-btr_search_sys_create(ulint hash_size)
+void btr_search_sys_create(ulint hash_size)
 {
 	/* Search System is divided into n parts.
 	Each part controls access to distinct set of hash buckets from
@@ -205,6 +254,12 @@ btr_search_sys_create(ulint hash_size)
 /** Frees the adaptive search system at a database shutdown. */
 void btr_search_sys_free()
 {
+  if (!btr_search_sys)
+  {
+    ut_ad(!btr_search_latches);
+    return;
+  }
+
   ut_ad(btr_search_sys);
   ut_ad(btr_search_latches);
 
@@ -378,9 +433,8 @@ void btr_search_disable()
 }
 
 /** Enable the adaptive hash search system.
-@param[in]	resize	Flag to indicate call during buf_pool_resize() */
-void
-btr_search_enable(bool resize)
+@param resize whether buf_pool_resize() is the caller */
+void btr_search_enable(bool resize)
 {
 	if (!resize) {
 		buf_pool_mutex_enter_all();
@@ -431,8 +485,8 @@ btr_search_info_update_hash(
 	ulint		n_unique;
 	int		cmp;
 
-	ut_ad(!rw_lock_own_flagged(btr_get_search_latch(index),
-				   RW_LOCK_FLAG_X | RW_LOCK_FLAG_S));
+	ut_ad(!btr_search_own_any(RW_LOCK_S));
+	ut_ad(!btr_search_own_any(RW_LOCK_X));
 
 	if (dict_index_is_ibuf(index)) {
 		/* So many deletes are performed on an insert buffer tree
@@ -536,17 +590,12 @@ block->n_hash_helps, n_fields, n_bytes, left_side are NOT protected by any
 semaphore, to save CPU time! Do not assume the fields are consistent.
 @return TRUE if building a (new) hash index on the block is recommended
 @param[in,out]	info	search info
-@param[in,out]	block	buffer block
-@param[in]	cursor	cursor */
+@param[in,out]	block	buffer block */
 static
-ibool
-btr_search_update_block_hash_info(
-	btr_search_t*		info,
-	buf_block_t*		block,
-	const btr_cur_t*	cursor)
+bool
+btr_search_update_block_hash_info(btr_search_t* info, buf_block_t* block)
 {
-	ut_ad(!rw_lock_own_flagged(btr_get_search_latch(cursor->index),
-				   RW_LOCK_FLAG_X | RW_LOCK_FLAG_S));
+	ut_ad(!btr_search_own_any());
 	ut_ad(rw_lock_own_flagged(&block->lock,
 				  RW_LOCK_FLAG_X | RW_LOCK_FLAG_S));
 
@@ -593,11 +642,11 @@ btr_search_update_block_hash_info(
 
 			/* Build a new hash index on the page */
 
-			return(TRUE);
+			return(true);
 		}
 	}
 
-	return(FALSE);
+	return(false);
 }
 
 /** Updates a hash node reference when it has been unsuccessfully used in a
@@ -637,8 +686,8 @@ btr_search_update_hash_ref(
 		return;
 	}
 
-	ut_ad(block->page.id.space() == index->space);
-	ut_a(index == cursor->index);
+	ut_ad(block->page.id.space() == index->table->space_id);
+	ut_ad(index == cursor->index);
 	ut_ad(!dict_index_is_ibuf(index));
 	rw_lock_t* const latch = btr_get_search_latch(index);
 	rw_lock_x_lock(latch);
@@ -661,7 +710,8 @@ btr_search_update_hash_ref(
 
 		ulint fold = rec_fold(
 			rec,
-			rec_get_offsets(rec, index, offsets_, true,
+			rec_get_offsets(rec, index, offsets_,
+					index->n_core_fields,
 					ULINT_UNDEFINED, &heap),
 			block->curr_n_fields,
 			block->curr_n_bytes, index->id);
@@ -679,56 +729,6 @@ func_exit:
 	rw_lock_x_unlock(latch);
 }
 
-/** Updates the search info.
-@param[in,out]	info	search info
-@param[in]	cursor	cursor which was just positioned */
-void
-btr_search_info_update_slow(
-	btr_search_t*	info,
-	btr_cur_t*	cursor)
-{
-	buf_block_t*	block;
-	ibool		build_index;
-
-	ut_ad(!rw_lock_own_flagged(btr_get_search_latch(cursor->index),
-				   RW_LOCK_FLAG_X | RW_LOCK_FLAG_S));
-
-	block = btr_cur_get_block(cursor);
-
-	/* NOTE that the following two function calls do NOT protect
-	info or block->n_fields etc. with any semaphore, to save CPU time!
-	We cannot assume the fields are consistent when we return from
-	those functions! */
-
-	btr_search_info_update_hash(info, cursor);
-
-	build_index = btr_search_update_block_hash_info(info, block, cursor);
-
-	if (build_index || (cursor->flag == BTR_CUR_HASH_FAIL)) {
-
-		btr_search_check_free_space_in_heap(cursor->index);
-	}
-
-	if (cursor->flag == BTR_CUR_HASH_FAIL) {
-		/* Update the hash node reference, if appropriate */
-
-#ifdef UNIV_SEARCH_PERF_STAT
-		btr_search_n_hash_fail++;
-#endif /* UNIV_SEARCH_PERF_STAT */
-		btr_search_update_hash_ref(info, block, cursor);
-	}
-
-	if (build_index) {
-		/* Note that since we did not protect block->n_fields etc.
-		with any semaphore, the values can be inconsistent. We have
-		to check inside the function call that they make sense. */
-		btr_search_build_page_hash_index(cursor->index, block,
-						 block->n_fields,
-						 block->n_bytes,
-						 block->left_side);
-	}
-}
-
 /** Checks if a guessed position for a tree cursor is right. Note that if
 mode is PAGE_CUR_LE, which is used in inserts, and the function returns
 TRUE, then cursor->up_match and cursor->low_match both have sensible values.
@@ -742,16 +742,14 @@ TRUE, then cursor->up_match and cursor->low_match both have sensible values.
 				previous record to check our guess!
 @param[in]	tuple		data tuple
 @param[in]	mode		PAGE_CUR_L, PAGE_CUR_LE, PAGE_CUR_G, PAGE_CUR_GE
-@param[in]	mtr		mini transaction
-@return TRUE if success */
+@return	whether a match was found */
 static
-ibool
+bool
 btr_search_check_guess(
 	btr_cur_t*	cursor,
-	ibool		can_only_compare_to_cursor_rec,
+	bool		can_only_compare_to_cursor_rec,
 	const dtuple_t*	tuple,
-	ulint		mode,
-	mtr_t*		mtr)
+	ulint		mode)
 {
 	rec_t*		rec;
 	ulint		n_unique;
@@ -772,7 +770,8 @@ btr_search_check_guess(
 
 	match = 0;
 
-	offsets = rec_get_offsets(rec, cursor->index, offsets, true,
+	offsets = rec_get_offsets(rec, cursor->index, offsets,
+				  cursor->index->n_core_fields,
 				  n_unique, &heap);
 	cmp = cmp_dtuple_rec_with_match(tuple, rec, offsets, &match);
 
@@ -813,11 +812,9 @@ btr_search_check_guess(
 	match = 0;
 
 	if ((mode == PAGE_CUR_G) || (mode == PAGE_CUR_GE)) {
-		rec_t*	prev_rec;
-
 		ut_ad(!page_rec_is_infimum(rec));
 
-		prev_rec = page_rec_get_prev(rec);
+		const rec_t* prev_rec = page_rec_get_prev(rec);
 
 		if (page_rec_is_infimum(prev_rec)) {
 			success = !page_has_prev(page_align(prev_rec));
@@ -825,7 +822,8 @@ btr_search_check_guess(
 		}
 
 		offsets = rec_get_offsets(prev_rec, cursor->index, offsets,
-					  true, n_unique, &heap);
+					  cursor->index->n_core_fields,
+					  n_unique, &heap);
 		cmp = cmp_dtuple_rec_with_match(
 			tuple, prev_rec, offsets, &match);
 		if (mode == PAGE_CUR_GE) {
@@ -833,14 +831,10 @@ btr_search_check_guess(
 		} else {
 			success = cmp >= 0;
 		}
-
-		goto exit_func;
 	} else {
-		rec_t*	next_rec;
-
 		ut_ad(!page_rec_is_supremum(rec));
 
-		next_rec = page_rec_get_next(rec);
+		const rec_t* next_rec = page_rec_get_next(rec);
 
 		if (page_rec_is_supremum(next_rec)) {
 			if (!page_has_next(page_align(next_rec))) {
@@ -852,7 +846,8 @@ btr_search_check_guess(
 		}
 
 		offsets = rec_get_offsets(next_rec, cursor->index, offsets,
-					  true, n_unique, &heap);
+					  cursor->index->n_core_fields,
+					  n_unique, &heap);
 		cmp = cmp_dtuple_rec_with_match(
 			tuple, next_rec, offsets, &match);
 		if (mode == PAGE_CUR_LE) {
@@ -900,12 +895,9 @@ both have sensible values.
 				we assume the caller uses his search latch
 				to protect the record!
 @param[out]	cursor		tree cursor
-@param[in]	has_search_latch
-				latch mode the caller currently has on
-				search system: RW_S/X_LATCH or 0
 @param[in]	mtr		mini transaction
-@return TRUE if succeeded */
-ibool
+@return whether the search succeeded */
+bool
 btr_search_guess_on_hash(
 	dict_index_t*	index,
 	btr_search_t*	info,
@@ -913,7 +905,6 @@ btr_search_guess_on_hash(
 	ulint		mode,
 	ulint		latch_mode,
 	btr_cur_t*	cursor,
-	ulint		has_search_latch,
 	mtr_t*		mtr)
 {
 	ulint		fold;
@@ -924,7 +915,7 @@ btr_search_guess_on_hash(
 #endif
 
 	if (!btr_search_enabled) {
-		return(FALSE);
+		return false;
 	}
 
 	ut_ad(index && info && tuple && cursor && mtr);
@@ -939,16 +930,14 @@ btr_search_guess_on_hash(
 	any latch here! */
 
 	if (info->n_hash_potential == 0) {
-
-		return(FALSE);
+		return false;
 	}
 
 	cursor->n_fields = info->n_fields;
 	cursor->n_bytes = info->n_bytes;
 
 	if (dtuple_get_n_fields(tuple) < btr_search_get_n_fields(cursor)) {
-
-		return(FALSE);
+		return false;
 	}
 
 	index_id = index->id;
@@ -961,65 +950,53 @@ btr_search_guess_on_hash(
 	cursor->fold = fold;
 	cursor->flag = BTR_CUR_HASH;
 
-	rw_lock_t* const latch = btr_get_search_latch(index);
+	rw_lock_t* ahi_latch = btr_get_search_latch(index);
+	const rec_t* rec;
 
-	if (!has_search_latch) {
-		rw_lock_s_lock(latch);
+		rw_lock_s_lock(ahi_latch);
 
-		if (!btr_search_enabled) {
-fail:
-			if (!has_search_latch) {
-				rw_lock_s_unlock(latch);
-			}
-			btr_search_failure(info, cursor);
-			return(FALSE);
-		}
+	if (!btr_search_enabled) {
+		goto fail;
 	}
 
-	ut_ad(rw_lock_own(latch, RW_LOCK_S));
-
-	const rec_t* rec = static_cast<const rec_t*>(
+	rec = static_cast<const rec_t*>(
 		ha_search_and_get_data(btr_get_search_table(index), fold));
 
 	if (!rec) {
-		goto fail;
+fail:
+		rw_lock_s_unlock(ahi_latch);
+
+		btr_search_failure(info, cursor);
+		return false;
 	}
 
 	buf_block_t*	block = buf_block_from_ahi(rec);
 
-	if (!has_search_latch) {
+	if (!buf_page_get_known_nowait(latch_mode, block, BUF_MAKE_YOUNG,
+	      __FILE__, __LINE__, mtr)) {
+		goto fail;
+	}
 
-		if (!buf_page_get_known_nowait(
-			latch_mode, block, BUF_MAKE_YOUNG,
-			__FILE__, __LINE__, mtr)) {
-			goto fail;
-		}
+	const bool fail = index != block->index
+		&& index_id == block->index->id;
+	ut_a(!fail || block->index->freed());
+	ut_ad(fail || !block->page.file_page_was_freed);
+	rw_lock_s_unlock(ahi_latch);
 
-		const bool fail = index != block->index
-			&& index_id == block->index->id;
-		ut_a(!fail || block->index->freed());
-		rw_lock_s_unlock(latch);
-
-		buf_block_dbg_add_level(block, SYNC_TREE_NODE_FROM_HASH);
-		if (UNIV_UNLIKELY(fail)) {
-			goto fail_and_release_page;
-		}
-	} else if (UNIV_UNLIKELY(index != block->index
-				 && index_id == block->index->id)) {
-		ut_a(block->index->freed());
+	buf_block_dbg_add_level(block, SYNC_TREE_NODE_FROM_HASH);
+	if (UNIV_UNLIKELY(fail)) {
 		goto fail_and_release_page;
 	}
 
 	if (buf_block_get_state(block) != BUF_BLOCK_FILE_PAGE) {
 
 		ut_ad(buf_block_get_state(block) == BUF_BLOCK_REMOVE_HASH);
+
 fail_and_release_page:
-		if (!has_search_latch) {
-			btr_leaf_page_release(block, latch_mode, mtr);
-		}
+		btr_leaf_page_release(block, latch_mode, mtr);
 
 		btr_search_failure(info, cursor);
-		return(FALSE);
+		return false;
 	}
 
 	ut_ad(page_rec_is_user_rec(rec));
@@ -1034,9 +1011,7 @@ fail_and_release_page:
 	record to determine if our guess for the cursor position is
 	right. */
 	if (index_id != btr_page_get_index_id(block->frame)
-	    || !btr_search_check_guess(cursor,
-				       has_search_latch,
-				       tuple, mode, mtr)) {
+	    || !btr_search_check_guess(cursor, 0, tuple, mode)) {
 		goto fail_and_release_page;
 	}
 
@@ -1052,7 +1027,7 @@ fail_and_release_page:
 	info->last_hash_succ = FALSE;
 
 	/* Currently, does not work if the following fails: */
-	ut_ad(!has_search_latch);
+	ut_ad(!ahi_latch);
 
 	btr_leaf_page_release(block, latch_mode, mtr);
 
@@ -1085,7 +1060,7 @@ fail_and_release_page:
 #ifdef UNIV_SEARCH_PERF_STAT
 	btr_search_n_succ++;
 #endif
-	if (!has_search_latch && buf_page_peek_if_too_old(&block->page)) {
+	if (buf_page_peek_if_too_old(&block->page)) {
 
 		buf_page_make_young(&block->page);
 	}
@@ -1098,7 +1073,7 @@ fail_and_release_page:
 		++buf_pool->stat.n_page_gets;
 	}
 
-	return(TRUE);
+	return true;
 }
 
 /** Drop any adaptive hash index entries that point to an index page.
@@ -1106,9 +1081,11 @@ fail_and_release_page:
 			index page for which we know that
 			block->buf_fix_count == 0 or it is an index page which
 			has already been removed from the buf_pool->page_hash
-			i.e.: it is in state BUF_BLOCK_REMOVE_HASH */
-void
-btr_search_drop_page_hash_index(buf_block_t* block)
+			i.e.: it is in state BUF_BLOCK_REMOVE_HASH
+@param[in]	garbage_collect	drop ahi only if the index is marked
+				as freed */
+void btr_search_drop_page_hash_index(buf_block_t* block,
+				     bool garbage_collect)
 {
 	ulint			n_fields;
 	ulint			n_bytes;
@@ -1125,9 +1102,8 @@ btr_search_drop_page_hash_index(buf_block_t* block)
 	rw_lock_t*		latch;
 
 retry:
-	/* This debug check uses a dirty read that could theoretically cause
-	false positives while buf_pool_clear_hash_index() is executing. */
-	assert_block_ahi_valid(block);
+	ut_ad(!btr_search_own_any(RW_LOCK_S));
+	ut_ad(!btr_search_own_any(RW_LOCK_X));
 
 	if (!block->index) {
 		return;
@@ -1152,15 +1128,21 @@ retry:
 		% btr_ahi_parts;
 	latch = btr_search_latches[ahi_slot];
 
+	rw_lock_s_lock(latch);
+
 	dict_index_t* index = block->index;
-	ut_ad(!btr_search_own_any(RW_LOCK_S));
-	ut_ad(!btr_search_own_any(RW_LOCK_X));
 
 	bool is_freed = index && index->freed();
 	if (is_freed) {
+		rw_lock_s_unlock(latch);
 		rw_lock_x_lock(latch);
-	} else {
-		rw_lock_s_lock(latch);
+		if (index != block->index) {
+			rw_lock_x_unlock(latch);
+			goto retry;
+		}
+	} else if (garbage_collect) {
+		rw_lock_s_unlock(latch);
+		return;
 	}
 
 	assert_block_ahi_valid(block);
@@ -1179,8 +1161,7 @@ retry:
 #endif
 	ut_ad(btr_search_enabled);
 
-	ut_ad(index->space == FIL_NULL
-	      || block->page.id.space() == index->space);
+	ut_ad(block->page.id.space() == index->table->space_id);
 	ut_a(index_id == index->id);
 	ut_ad(!dict_index_is_ibuf(index));
 
@@ -1208,6 +1189,9 @@ retry:
 
 	rec = page_get_infimum_rec(page);
 	rec = page_rec_get_next_low(rec, page_is_comp(page));
+	if (rec_is_metadata(rec, index)) {
+		rec = page_rec_get_next_low(rec, page_is_comp(page));
+	}
 
 	prev_fold = 0;
 
@@ -1216,7 +1200,7 @@ retry:
 
 	while (!page_rec_is_supremum(rec)) {
 		offsets = rec_get_offsets(
-			rec, index, offsets, true,
+			rec, index, offsets, index->n_core_fields,
 			btr_search_get_n_fields(n_fields, n_bytes),
 			&heap);
 		fold = rec_fold(rec, offsets, n_fields, n_bytes, index_id);
@@ -1341,6 +1325,7 @@ If index is non-NULL, this function checks if n_fields and n_bytes are
 sensible, and does not build a hash index if not.
 @param[in,out]	index		index for which to build.
 @param[in,out]	block		index page, s-/x- latched.
+@param[in,out]	ahi_latch	the adaptive search latch
 @param[in]	n_fields	hash this many full fields
 @param[in]	n_bytes		hash this many bytes of the next field
 @param[in]	left_side	hash for searches from left side */
@@ -1349,21 +1334,19 @@ void
 btr_search_build_page_hash_index(
 	dict_index_t*	index,
 	buf_block_t*	block,
+	rw_lock_t*	ahi_latch,
 	ulint		n_fields,
 	ulint		n_bytes,
 	ibool		left_side)
 {
-	hash_table_t*	table;
-	page_t*		page;
-	rec_t*		rec;
-	rec_t*		next_rec;
+	const rec_t*	rec;
+	const rec_t*	next_rec;
 	ulint		fold;
 	ulint		next_fold;
 	ulint		n_cached;
 	ulint		n_recs;
 	ulint*		folds;
-	rec_t**		recs;
-	ulint		i;
+	const rec_t**	recs;
 	mem_heap_t*	heap		= NULL;
 	rec_offs	offsets_[REC_OFFS_NORMAL_SIZE];
 	rec_offs*	offsets		= offsets_;
@@ -1376,8 +1359,9 @@ btr_search_build_page_hash_index(
 	}
 
 	rec_offs_init(offsets_);
+	ut_ad(ahi_latch == btr_get_search_latch(index));
 	ut_ad(index);
-	ut_ad(block->page.id.space() == index->space);
+	ut_ad(block->page.id.space() == index->table->space_id);
 	ut_ad(!dict_index_is_ibuf(index));
 	ut_ad(page_is_leaf(block->frame));
 
@@ -1385,24 +1369,21 @@ btr_search_build_page_hash_index(
 				  RW_LOCK_FLAG_X | RW_LOCK_FLAG_S));
 	ut_ad(block->page.id.page_no() >= 3);
 
-	rw_lock_t* const latch = btr_get_search_latch(index);
-	rw_lock_s_lock(latch);
+	rw_lock_s_lock(ahi_latch);
 
-	if (!btr_search_enabled) {
-		rw_lock_s_unlock(latch);
+	const bool enabled = btr_search_enabled;
+	const bool rebuild = enabled && block->index
+		&& (block->curr_n_fields != n_fields
+		    || block->curr_n_bytes != n_bytes
+		    || block->curr_left_side != left_side);
+
+	rw_lock_s_unlock(ahi_latch);
+
+	if (!enabled) {
 		return;
 	}
 
-	table = btr_get_search_table(index);
-	page = buf_block_get_frame(block);
-
-	const bool must_drop = block->index
-		&& ((block->curr_n_fields != n_fields)
-		    || (block->curr_n_bytes != n_bytes)
-		    || (block->curr_left_side != left_side));
-	rw_lock_s_unlock(latch);
-
-	if (must_drop) {
+	if (rebuild) {
 		btr_search_drop_page_hash_index(block);
 	}
 
@@ -1418,6 +1399,7 @@ btr_search_build_page_hash_index(
 		return;
 	}
 
+	page_t*		page	= buf_block_get_frame(block);
 	n_recs = page_get_n_recs(page);
 
 	if (n_recs == 0) {
@@ -1425,20 +1407,26 @@ btr_search_build_page_hash_index(
 		return;
 	}
 
+	rec = page_rec_get_next_const(page_get_infimum_rec(page));
+
+	if (rec_is_metadata(rec, index)) {
+		rec = page_rec_get_next_const(rec);
+		if (!--n_recs) return;
+	}
+
 	/* Calculate and cache fold values and corresponding records into
 	an array for fast insertion to the hash index */
 
-	folds = (ulint*) ut_malloc_nokey(n_recs * sizeof(ulint));
-	recs = (rec_t**) ut_malloc_nokey(n_recs * sizeof(rec_t*));
+	folds = static_cast<ulint*>(ut_malloc_nokey(n_recs * sizeof *folds));
+	recs = static_cast<const rec_t**>(
+		ut_malloc_nokey(n_recs * sizeof *recs));
 
 	n_cached = 0;
 
 	ut_a(index->id == btr_page_get_index_id(page));
 
-	rec = page_rec_get_next(page_get_infimum_rec(page));
-
 	offsets = rec_get_offsets(
-		rec, index, offsets, true,
+		rec, index, offsets, index->n_core_fields,
 		btr_search_get_n_fields(n_fields, n_bytes),
 		&heap);
 	ut_ad(page_rec_is_supremum(rec)
@@ -1454,7 +1442,7 @@ btr_search_build_page_hash_index(
 	}
 
 	for (;;) {
-		next_rec = page_rec_get_next(rec);
+		next_rec = page_rec_get_next_const(rec);
 
 		if (page_rec_is_supremum(next_rec)) {
 
@@ -1469,7 +1457,7 @@ btr_search_build_page_hash_index(
 		}
 
 		offsets = rec_get_offsets(
-			next_rec, index, offsets, true,
+			next_rec, index, offsets, index->n_core_fields,
 			btr_search_get_n_fields(n_fields, n_bytes), &heap);
 		next_fold = rec_fold(next_rec, offsets, n_fields,
 				     n_bytes, index->id);
@@ -1495,19 +1483,12 @@ btr_search_build_page_hash_index(
 
 	btr_search_check_free_space_in_heap(index);
 
-	rw_lock_x_lock(latch);
+	rw_lock_x_lock(ahi_latch);
 
 	if (!btr_search_enabled) {
 		goto exit_func;
 	}
 
-	table = btr_get_search_table(index);
-	if (block->index && ((block->curr_n_fields != n_fields)
-			     || (block->curr_n_bytes != n_bytes)
-			     || (block->curr_left_side != left_side))) {
-		goto exit_func;
-	}
-
 	/* This counter is decremented every time we drop page
 	hash index entries and is incremented here. Since we can
 	rebuild hash index for a page that is already hashed, we
@@ -1516,6 +1497,10 @@ btr_search_build_page_hash_index(
 	if (!block->index) {
 		assert_block_ahi_empty(block);
 		index->search_info->ref_count++;
+	} else if (block->curr_n_fields != n_fields
+		   || block->curr_n_bytes != n_bytes
+		   || block->curr_left_side != left_side) {
+		goto exit_func;
 	}
 
 	block->n_hash_helps = 0;
@@ -1525,16 +1510,18 @@ btr_search_build_page_hash_index(
 	block->curr_left_side = unsigned(left_side);
 	block->index = index;
 
-	for (i = 0; i < n_cached; i++) {
-
-		ha_insert_for_fold(table, folds[i], block, recs[i]);
+	{
+		hash_table_t*	table = btr_get_search_table(index);
+		for (ulint i = 0; i < n_cached; i++) {
+			ha_insert_for_fold(table, folds[i], block, recs[i]);
+		}
 	}
 
 	MONITOR_INC(MONITOR_ADAPTIVE_HASH_PAGE_ADDED);
 	MONITOR_INC_VALUE(MONITOR_ADAPTIVE_HASH_ROW_ADDED, n_cached);
 exit_func:
 	assert_block_ahi_valid(block);
-	rw_lock_x_unlock(latch);
+	rw_lock_x_unlock(ahi_latch);
 
 	ut_free(folds);
 	ut_free(recs);
@@ -1543,50 +1530,100 @@ exit_func:
 	}
 }
 
-/** Moves or deletes hash entries for moved records. If new_page is already
-hashed, then the hash index for page, if any, is dropped. If new_page is not
-hashed, and page is hashed, then a new hash index is built to new_page with the
-same parameters as page (this often happens when a page is split).
-@param[in,out]	new_block	records are copied to this page.
-@param[in,out]	block		index page from which record are copied, and the
-				copied records will be deleted from this page.
-@param[in,out]	index		record descriptor */
+/** Updates the search info.
+@param[in,out]	info	search info
+@param[in,out]	cursor	cursor which was just positioned */
+void
+btr_search_info_update_slow(btr_search_t* info, btr_cur_t* cursor)
+{
+	rw_lock_t*	ahi_latch = btr_get_search_latch(cursor->index);
+
+	ut_ad(!rw_lock_own_flagged(ahi_latch,
+				   RW_LOCK_FLAG_X | RW_LOCK_FLAG_S));
+
+	buf_block_t*	block = btr_cur_get_block(cursor);
+
+	/* NOTE that the following two function calls do NOT protect
+	info or block->n_fields etc. with any semaphore, to save CPU time!
+	We cannot assume the fields are consistent when we return from
+	those functions! */
+
+	btr_search_info_update_hash(info, cursor);
+
+	bool build_index = btr_search_update_block_hash_info(info, block);
+
+	if (build_index || (cursor->flag == BTR_CUR_HASH_FAIL)) {
+
+		btr_search_check_free_space_in_heap(cursor->index);
+	}
+
+	if (cursor->flag == BTR_CUR_HASH_FAIL) {
+		/* Update the hash node reference, if appropriate */
+
+#ifdef UNIV_SEARCH_PERF_STAT
+		btr_search_n_hash_fail++;
+#endif /* UNIV_SEARCH_PERF_STAT */
+
+		btr_search_update_hash_ref(info, block, cursor);
+	}
+
+	if (build_index) {
+		/* Note that since we did not protect block->n_fields etc.
+		with any semaphore, the values can be inconsistent. We have
+		to check inside the function call that they make sense. */
+		btr_search_build_page_hash_index(cursor->index, block,
+						 ahi_latch,
+						 block->n_fields,
+						 block->n_bytes,
+						 block->left_side);
+	}
+}
+
+/** Move or delete hash entries for moved records, usually in a page split.
+If new_block is already hashed, then any hash index for block is dropped.
+If new_block is not hashed, and block is hashed, then a new hash index is
+built to new_block with the same parameters as block.
+@param[in,out]	new_block	destination page
+@param[in,out]	block		source page (subject to deletion later) */
 void
 btr_search_move_or_delete_hash_entries(
 	buf_block_t*	new_block,
-	buf_block_t*	block,
-	dict_index_t*	index)
+	buf_block_t*	block)
 {
-#ifdef MYSQL_INDEX_DISABLE_AHI
-	if (index->disable_ahi) return;
-#endif
+	ut_ad(rw_lock_own(&(block->lock), RW_LOCK_X));
+	ut_ad(rw_lock_own(&(new_block->lock), RW_LOCK_X));
+
 	if (!btr_search_enabled) {
 		return;
 	}
 
-	ut_ad(rw_lock_own(&(block->lock), RW_LOCK_X));
-	ut_ad(rw_lock_own(&(new_block->lock), RW_LOCK_X));
-
-	rw_lock_t* const latch = btr_get_search_latch(index);
-	rw_lock_s_lock(latch);
-
-	ut_a(!new_block->index || new_block->index == index);
-	ut_a(!block->index || block->index->id == index->id);
-	ut_ad(!(new_block->index || block->index)
-	      || !dict_index_is_ibuf(index));
+	dict_index_t* index = block->index;
+	if (!index) {
+		index = new_block->index;
+	} else {
+		ut_ad(!new_block->index || index == new_block->index);
+	}
 	assert_block_ahi_valid(block);
 	assert_block_ahi_valid(new_block);
 
+	rw_lock_t* ahi_latch = index ? btr_get_search_latch(index) : NULL;
+
 	if (new_block->index) {
 drop_exit:
-		rw_lock_s_unlock(latch);
 		btr_search_drop_page_hash_index(block);
 		return;
 	}
 
+	if (!index) {
+		return;
+	}
+
+	rw_lock_s_lock(ahi_latch);
+
 	if (block->index) {
 
 		if (block->index != index) {
+			rw_lock_s_unlock(ahi_latch);
 			goto drop_exit;
 		}
 
@@ -1598,27 +1635,26 @@ drop_exit:
 		new_block->n_bytes = block->curr_n_bytes;
 		new_block->left_side = left_side;
 
-		rw_lock_s_unlock(latch);
+		rw_lock_s_unlock(ahi_latch);
 
 		ut_a(n_fields > 0 || n_bytes > 0);
 
 		btr_search_build_page_hash_index(
-			index, new_block, n_fields, n_bytes, left_side);
+			index, new_block, ahi_latch,
+			n_fields, n_bytes, left_side);
 		ut_ad(n_fields == block->curr_n_fields);
 		ut_ad(n_bytes == block->curr_n_bytes);
 		ut_ad(left_side == block->curr_left_side);
 		return;
 	}
-	rw_lock_s_unlock(latch);
+	rw_lock_s_unlock(ahi_latch);
 }
 
 /** Updates the page hash index when a single record is deleted from a page.
 @param[in]	cursor	cursor which was positioned on the record to delete
 			using btr_cur_search_, the record is not yet deleted.*/
-void
-btr_search_update_hash_on_delete(btr_cur_t* cursor)
+void btr_search_update_hash_on_delete(btr_cur_t* cursor)
 {
-	hash_table_t*	table;
 	buf_block_t*	block;
 	const rec_t*	rec;
 	ulint		fold;
@@ -1654,58 +1690,60 @@ btr_search_update_hash_on_delete(btr_cur_t* cursor)
 		return;
 	}
 
-	ut_ad(block->page.id.space() == index->space);
+	ut_ad(block->page.id.space() == index->table->space_id);
 	ut_a(index == cursor->index);
 	ut_a(block->curr_n_fields > 0 || block->curr_n_bytes > 0);
 	ut_ad(!dict_index_is_ibuf(index));
 
 	rec = btr_cur_get_rec(cursor);
 
-	fold = rec_fold(rec, rec_get_offsets(rec, index, offsets_, true,
+	fold = rec_fold(rec, rec_get_offsets(rec, index, offsets_,
+					     index->n_core_fields,
 					     ULINT_UNDEFINED, &heap),
 			block->curr_n_fields, block->curr_n_bytes, index->id);
 	if (UNIV_LIKELY_NULL(heap)) {
 		mem_heap_free(heap);
 	}
 
-	rw_lock_t* const latch = btr_get_search_latch(index);
-	rw_lock_x_lock(latch);
+	rw_lock_t* ahi_latch = btr_get_search_latch(index);
+
+	rw_lock_x_lock(ahi_latch);
 	assert_block_ahi_valid(block);
 
-	if (!btr_search_enabled) {
-		rw_lock_x_unlock(latch);
-		return;
-	}
+	if (btr_search_enabled) {
+		hash_table_t* table = btr_get_search_table(index);
+		if (block->index) {
+			ut_a(block->index == index);
 
-	table = btr_get_search_table(index);
-	if (block->index) {
-		ut_a(block->index == index);
+			if (ha_search_and_delete_if_found(table, fold, rec)) {
+				MONITOR_INC(MONITOR_ADAPTIVE_HASH_ROW_REMOVED);
+			} else {
+				MONITOR_INC(MONITOR_ADAPTIVE_HASH_ROW_REMOVE_NOT_FOUND);
+			}
 
-		if (ha_search_and_delete_if_found(table, fold, rec)) {
-			MONITOR_INC(MONITOR_ADAPTIVE_HASH_ROW_REMOVED);
-		} else {
-			MONITOR_INC(
-				MONITOR_ADAPTIVE_HASH_ROW_REMOVE_NOT_FOUND);
+			assert_block_ahi_valid(block);
 		}
-
-		assert_block_ahi_valid(block);
 	}
 
-	rw_lock_x_unlock(latch);
+	rw_lock_x_unlock(ahi_latch);
 }
 
 /** Updates the page hash index when a single record is inserted on a page.
 @param[in]	cursor	cursor which was positioned to the place to insert
 			using btr_cur_search_, and the new record has been
-			inserted next to the cursor. */
+			inserted next to the cursor.
+@param[in]	ahi_latch	the adaptive hash index latch */
 void
-btr_search_update_hash_node_on_insert(btr_cur_t* cursor)
+btr_search_update_hash_node_on_insert(btr_cur_t* cursor, rw_lock_t* ahi_latch)
 {
 	hash_table_t*	table;
 	buf_block_t*	block;
 	dict_index_t*	index;
 	rec_t*		rec;
 
+	ut_ad(ahi_latch == btr_get_search_latch(cursor->index));
+	ut_ad(!btr_search_own_any(RW_LOCK_S));
+	ut_ad(!btr_search_own_any(RW_LOCK_X));
 #ifdef MYSQL_INDEX_DISABLE_AHI
 	if (cursor->index->disable_ahi) return;
 #endif
@@ -1734,9 +1772,7 @@ btr_search_update_hash_node_on_insert(btr_cur_t* cursor)
 
 	ut_a(cursor->index == index);
 	ut_ad(!dict_index_is_ibuf(index));
-
-	rw_lock_t* const latch = btr_get_search_latch(index);
-	rw_lock_x_lock(latch);
+	rw_lock_x_lock(ahi_latch);
 
 	if (!block->index || !btr_search_enabled) {
 
@@ -1760,10 +1796,11 @@ btr_search_update_hash_node_on_insert(btr_cur_t* cursor)
 
 func_exit:
 		assert_block_ahi_valid(block);
-		rw_lock_x_unlock(latch);
+		rw_lock_x_unlock(ahi_latch);
 	} else {
-		rw_lock_x_unlock(latch);
-		btr_search_update_hash_on_insert(cursor);
+		rw_lock_x_unlock(ahi_latch);
+
+		btr_search_update_hash_on_insert(cursor, ahi_latch);
 	}
 }
 
@@ -1771,9 +1808,10 @@ func_exit:
 @param[in,out]	cursor		cursor which was positioned to the
 				place to insert using btr_cur_search_...,
 				and the new record has been inserted next
-				to the cursor */
+				to the cursor
+@param[in]	ahi_latch	the adaptive hash index latch */
 void
-btr_search_update_hash_on_insert(btr_cur_t* cursor)
+btr_search_update_hash_on_insert(btr_cur_t* cursor, rw_lock_t* ahi_latch)
 {
 	buf_block_t*	block;
 	dict_index_t*	index;
@@ -1790,7 +1828,10 @@ btr_search_update_hash_on_insert(btr_cur_t* cursor)
 	rec_offs*	offsets		= offsets_;
 	rec_offs_init(offsets_);
 
+	ut_ad(ahi_latch == btr_get_search_latch(cursor->index));
 	ut_ad(page_is_leaf(btr_cur_get_page(cursor)));
+	ut_ad(!btr_search_own_any(RW_LOCK_S));
+	ut_ad(!btr_search_own_any(RW_LOCK_X));
 #ifdef MYSQL_INDEX_DISABLE_AHI
 	if (cursor->index->disable_ahi) return;
 #endif
@@ -1810,7 +1851,7 @@ btr_search_update_hash_on_insert(btr_cur_t* cursor)
 		return;
 	}
 
-	ut_ad(block->page.id.space() == index->space);
+	ut_ad(block->page.id.space() == index->table->space_id);
 	btr_search_check_free_space_in_heap(index);
 
 	rec = btr_cur_get_rec(cursor);
@@ -1834,32 +1875,32 @@ btr_search_update_hash_on_insert(btr_cur_t* cursor)
 	ins_rec = page_rec_get_next_const(rec);
 	next_rec = page_rec_get_next_const(ins_rec);
 
-	offsets = rec_get_offsets(ins_rec, index, offsets, true,
+	offsets = rec_get_offsets(ins_rec, index, offsets,
+				  index->n_core_fields,
 				  ULINT_UNDEFINED, &heap);
 	ins_fold = rec_fold(ins_rec, offsets, n_fields, n_bytes, index->id);
 
 	if (!page_rec_is_supremum(next_rec)) {
 		offsets = rec_get_offsets(
-			next_rec, index, offsets, true,
+			next_rec, index, offsets, index->n_core_fields,
 			btr_search_get_n_fields(n_fields, n_bytes), &heap);
 		next_fold = rec_fold(next_rec, offsets, n_fields,
 				     n_bytes, index->id);
 	}
 
-	rw_lock_t* const latch = btr_get_search_latch(index);
-	/* We must not look up "table" before acquiring the latch. */
+	/* We must not look up "table" before acquiring ahi_latch. */
 	hash_table_t* table = NULL;
 	bool locked = false;
 
-	if (!page_rec_is_infimum(rec)) {
+	if (!page_rec_is_infimum(rec) && !rec_is_metadata(rec, index)) {
 		offsets = rec_get_offsets(
-			rec, index, offsets, true,
+			rec, index, offsets, index->n_core_fields,
 			btr_search_get_n_fields(n_fields, n_bytes), &heap);
 		fold = rec_fold(rec, offsets, n_fields, n_bytes, index->id);
 	} else {
-		locked = left_side;
-		if (locked) {
-			rw_lock_x_lock(latch);
+		if (left_side) {
+			locked = true;
+			rw_lock_x_lock(ahi_latch);
 
 			if (!btr_search_enabled || !block->index) {
 				goto function_exit;
@@ -1876,7 +1917,7 @@ btr_search_update_hash_on_insert(btr_cur_t* cursor)
 
 		if (!locked) {
 			locked = true;
-			rw_lock_x_lock(latch);
+			rw_lock_x_lock(ahi_latch);
 
 			if (!btr_search_enabled || !block->index) {
 				goto function_exit;
@@ -1898,7 +1939,7 @@ check_next_rec:
 		if (!left_side) {
 			if (!locked) {
 				locked = true;
-				rw_lock_x_lock(latch);
+				rw_lock_x_lock(ahi_latch);
 
 				if (!btr_search_enabled || !block->index) {
 					goto function_exit;
@@ -1916,7 +1957,7 @@ check_next_rec:
 	if (ins_fold != next_fold) {
 		if (!locked) {
 			locked = true;
-			rw_lock_x_lock(latch);
+			rw_lock_x_lock(ahi_latch);
 
 			if (!btr_search_enabled || !block->index) {
 				goto function_exit;
@@ -1937,8 +1978,9 @@ function_exit:
 		mem_heap_free(heap);
 	}
 	if (locked) {
-		rw_lock_x_unlock(latch);
+		rw_lock_x_unlock(ahi_latch);
 	}
+	ut_ad(!rw_lock_own(ahi_latch, RW_LOCK_X));
 }
 
 #if defined UNIV_AHI_DEBUG || defined UNIV_DEBUG
@@ -2052,12 +2094,14 @@ btr_search_hash_table_validate(ulint hash_table_id)
 			}
 
 			ut_ad(!dict_index_is_ibuf(block->index));
-			ut_ad(block->page.id.space() == block->index->space);
+			ut_ad(block->page.id.space()
+			      == block->index->table->space_id);
 
 			page_index_id = btr_page_get_index_id(block->frame);
 
 			offsets = rec_get_offsets(
-				node->data, block->index, offsets, true,
+				node->data, block->index, offsets,
+				block->index->n_core_fields,
 				btr_search_get_n_fields(block->curr_n_fields,
 							block->curr_n_bytes),
 				&heap);
@@ -2163,5 +2207,22 @@ btr_search_validate()
 	return(true);
 }
 
+#ifdef UNIV_DEBUG
+bool btr_search_check_marked_free_index(const buf_block_t *block)
+{
+  const index_id_t index_id= btr_page_get_index_id(block->frame);
+
+  rw_lock_t *ahi_latch= btr_get_search_latch(
+    index_id, block->page.id.space());
+
+  rw_lock_s_lock(ahi_latch);
+
+  bool is_freed= block->index && block->index->freed();
+
+  rw_lock_s_unlock(ahi_latch);
+
+  return is_freed;
+}
+#endif /* UNIV_DEBUG */
 #endif /* defined UNIV_AHI_DEBUG || defined UNIV_DEBUG */
 #endif /* BTR_CUR_HASH_ADAPT */
